import warnings
from pathlib import Path

from nhssynth.utils.constants import TIME


def check_ending(fn: str, ending=".csv") -> str:
    return fn if fn.endswith(ending) else fn + ending


def format_io(
    fn_in: str,
    fn_out: str,
    fn_metadata: str,
    dir_data: str,
    run_name=TIME,
    dir_exp="experiments",
) -> tuple[Path, Path, Path, Path, Path]:
    # ensure .csv ending consistency
    fn_in, fn_out, fn_metadata = check_ending(fn_in), check_ending(fn_out), check_ending(fn_metadata, ending=".yaml")

    dir_data = Path(dir_data)

    # check if `fn_out` and `fn_metadata` are given as suffixes (start with an underscore) to append to `fn_in`
    # if not assume it is a name in its own right
    if fn_out[0] == "_":
        fn_out = check_ending(fn_in[:-4] + fn_out)
    if fn_metadata[0] == "_":
        fn_metadata = check_ending(fn_in[:-4] + fn_metadata, ending=".yaml")

    if "/" in fn_in:
        fn_in = Path(fn_in)
        warnings.warn(
            f"\nUsing the path supplied to `--input-file` appended to `--dir`, i.e. attempting to read data from {dir_data / fn_in},\nto avoid this warning, specify the path using `--dir` and only the name using `--input-file`\ne.g. `... --dir {(dir_data / fn_in).parent} --input-file {fn_in.name} ...`"
        )

    # generate timestamped experiment folder
    dir_exp = Path(dir_exp) / run_name

    if "/" in fn_out:
        fn_out = Path(fn_out).name
        warnings.warn(
            f"\nPaths are not supported via `--output-file`, using the name part instead, i.e. attempting to write data to {dir_exp / fn_out}"
        )

    if "/" in fn_metadata:
        fn_metadata = Path(fn_metadata)
        warnings.warn(
            f"\nUsing the path supplied to `--metadata` appended to `--dir`, i.e. attempting to read data from {dir_data / fn_metadata},\nto avoid this warning, specify the path using `--dir` and only the name using `--metadata`\ne.g. `... --dir {(dir_data / fn_metadata).parent} --metadata {fn_metadata.name} ...`"
        )

    return dir_data / fn_in, dir_exp / fn_out, dir_data / fn_metadata, dir_exp / fn_metadata, dir_exp



import warnings

import pandas as pd
from nhssynth.modules.dataloader.utils import *
from rdt import HyperTransformer
from rdt.transformers import *
from sdv.metadata import SingleTableMetadata
from sdv.single_table import TVAESynthesizer


def get_transformer(d: dict):
    transformer_data = d.get("transformer")
    if isinstance(transformer_data, dict):
        transformer_name = transformer_data.pop("name", None)
        return eval(transformer_name)(**transformer_data) if transformer_name else None
    elif transformer_data:
        return eval(transformer_data)()
    else:
        return None


def instantiate_synthesizer(sdtypes, transformers, data, allow_null_transformers: bool):
    if all(sdtypes.values()):
        metadata = SingleTableMetadata.load_from_dict({"columns": sdtypes})
    else:
        warnings.warn(
            f"Incomplete metadata, detecting missing `sdtype`s for column(s): {[k for k, v in sdtypes.items() if not v]} automatically...",
            UserWarning,
        )
        metadata = SingleTableMetadata()
        metadata.detect_from_dataframe(data)
        for column_name, values in sdtypes.items():
            if values:
                metadata.update_column(column_name=column_name, **values)
    if not all(transformers.values()) and not allow_null_transformers:
        warnings.warn(
            f"Incomplete metadata, detecting missing `transformers`s for column(s): {[k for k, v in transformers.items() if not v]} automatically...",
            UserWarning,
        )
    # TODO support choice of synthesizer to base this on (not sure how much it matters),
    # synthesizer on some level must form the basis upon which the transformers are inferred from the metadata
    synthesizer = TVAESynthesizer(metadata)
    synthesizer.auto_assign_transformers(data)
    synthesizer.update_transformers(
        transformers if allow_null_transformers else {k: v for k, v in transformers.items() if v}
    )
    return metadata, synthesizer


def instantiate_hypertransformer(sdtypes, transformers, data, allow_null_transformers: bool) -> HyperTransformer:
    ht = HyperTransformer()
    if all(sdtypes.values()) and (all(transformers.values()) or allow_null_transformers):
        ht.set_config(
            config={
                "sdtypes": {k: v["sdtype"] for k, v in sdtypes.items()},
                "transformers": transformers,
            }
        )
    else:
        warnings.warn(
            f"Incomplete metadata, detecting missing{(' `sdtype`s for column(s): ' + str([k for k, v in sdtypes.items() if not v])) if not all(sdtypes.values()) else ''}{(' `transformer`s for column(s): ' + str([k for k, v in transformers.items() if not v])) if not all(transformers.values()) and not allow_null_transformers else ''} automatically...",
            UserWarning,
        )
        ht.detect_initial_config(data)
        ht.update_sdtypes({k: v["sdtype"] for k, v in sdtypes.items() if v})
        ht.update_transformers(
            transformers if allow_null_transformers else {k: v for k, v in transformers.items() if v}
        )
    return ht


# TODO maybe there is a cleaner way of doing this functionality without having the empty dict space for unspecified cols in the data?
def instantiate_metatransformer(
    metadata: dict, data: pd.DataFrame, sdv_workflow: bool, allow_null_transformers: bool
) -> tuple[dict[str, dict], SingleTableMetadata]:

    sdtypes = {cn: filter_inner_dict(cd, {"dtype", "transformer"}) for cn, cd in metadata.items()}
    transformers = {cn: get_transformer(cd) for cn, cd in metadata.items()}
    if sdv_workflow:
        metatransformer = instantiate_synthesizer(sdtypes, transformers, data, allow_null_transformers)
    else:
        metatransformer = instantiate_hypertransformer(sdtypes, transformers, data, allow_null_transformers)
    return metatransformer


def apply_transformer(metatransformer, typed_input, sdv_workflow: bool):
    if sdv_workflow:
        pass
    else:
        pass



import numpy as np
from sdv.evaluation import evaluate
import gower
import pandas as pd

from sdv.metrics.tabular import NumericalMLP, CategoricalSVM

# Distributional metrics - Check distribution differences between synthetic & original dataset as well as how
# Easy it is to discriminate them i.e. svc detection
def distribution_metrics(
    gower_bool,
    distributional_metrics,
    data_supp,
    synthetic_supp,
    categorical_columns,
    continuous_columns,
    saving_filepath="metrics/",
    pre_proc_method="GMM",
):

    # Define lists to contain the metrics achieved

    no_metrics = len(distributional_metrics)
    metrics = []

    # Need these in same column order

    synthetic_supp = synthetic_supp[data_supp.columns]

    # Now categorical columns need to be converted to objects as SDV infers data
    # types from the fields and integers/floats are treated as numerical not categorical

    synthetic_supp[categorical_columns] = synthetic_supp[categorical_columns].astype(
        object
    )
    data_supp[categorical_columns] = data_supp[categorical_columns].astype(object)

    evals = evaluate(
        synthetic_supp, data_supp, metrics=distributional_metrics, aggregate=False
    )

    # evals is a pandas dataframe of metrics - if we want to add a gower metric then we can
    # save this separately

    metrics = np.array(evals["raw_score"])

    if gower_bool == True:

        # Find the gower distance
        metrics = np.append(
            metrics, np.mean(gower.gower_matrix(data_supp, synthetic_supp))
        )

        metrics = pd.DataFrame(
            data=[metrics], columns=(distributional_metrics + ["Gower"])
        )

    else:

        metrics = pd.DataFrame(data=[metrics], columns=(distributional_metrics))

    # Save these metrics into a pandas dataframe - if the user wants to

    if saving_filepath != None:

        metrics.to_csv(
            "{}Metrics_SynthVAE_{}.csv".format(saving_filepath, pre_proc_method)
        )

    return metrics


# Build in some privacy metrics from SDV - TO DO!!!


def privacy_metrics(
    private_variable,
    data_supp,
    synthetic_supp,
    categorical_columns,
    continuous_columns,
    saving_filepath=None,
    pre_proc_method="GMM",
):

    if private_variable in continuous_columns:

        continuous_columns = [
            column for column in continuous_columns if column != private_variable
        ]

        mlp_priv = NumericalMLP.compute(
            data_supp.fillna(0),
            synthetic_supp.fillna(0),
            key_fields=(continuous_columns),
            sensitive_fields=[private_variable],
        )

        return mlp_priv

    elif private_variable in categorical_columns:

        categorical_columns = [
            column for column in categorical_columns if column != private_variable
        ]

        svm_priv = CategoricalSVM.compute(
            data_supp.fillna(0),
            synthetic_supp.fillna(0),
            key_fields=(categorical_columns),
            sensitive_fields=[private_variable],
        )

        return svm_priv


# Build in some fairness metrics (will have to find a library/code these ourselves) - TO DO!!!


def fairness_metrics(
    user_metrics,
    data_supp,
    synthetic_supp,
    categorical_columns,
    continuous_columns,
    saving_filepath=None,
    pre_proc_method="GMM",
):

    return None



import argparse

from nhssynth.cli.module_setup import *
from nhssynth.utils.constants import TIME


def add_module_subparser(subparsers: argparse._SubParsersAction, name: str):
    parser = subparsers.add_parser(
        name=name,
        description=MODULE_MAP[name].description,
        help=MODULE_MAP[name].help,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    MODULE_MAP[name].args_func(parser)
    parser.set_defaults(func=MODULE_MAP[name].func)
    return parser


def add_top_level_args(parser: argparse.ArgumentParser):
    parser.add_argument(
        "-rn",
        "--run-name",
        default=TIME,
        help=f"name the run to affect logging, config and outputs, defaults to current time, i.e. `{TIME}`",
    )
    parser.add_argument(
        "-sc", "--save-config", action="store_true", help="provide this flag to save the config provided via the cli"
    )
    parser.add_argument(
        "-scp",
        "--save-config-path",
        help="where to save the config when `-sc` is provided, defaults to `config/dump_<RUN_NAME>.yaml`",
    )
    parser.add_argument("-s", "--seed", help="specify a seed for reproducibility")



import argparse

from nhssynth.cli.module_arguments import *
from nhssynth.modules import dataloader, evaluation, model, plotting, structure


class ModuleConfig:
    def __init__(self, func, args_func, description: str, help: str):
        self.func = func
        self.args_func = args_func
        self.description = description
        self.help = help


def run_pipeline(args: argparse.Namespace):
    print("Running full pipeline...")
    dataloader.run(args)
    structure.run(args)
    model.run(args)
    evaluation.run(args)
    plotting.run(args)


def config():
    raise "This function should never be called"


# TODO I wonder if it is possible to add a flag to ModuleConfig
# saying whether the module should be included in a full run,
# and construct the functions in module_arguments accordingly
MODULE_MAP = {
    "pipeline": ModuleConfig(
        run_pipeline,
        add_all_module_args,
        "Run an automatically configured module or set of modules specified by a config file in `config/`. Note that you can override parts of the configuration on the fly via the usual CLI flags.",
        "run the full pipeline",
    ),
    "config": ModuleConfig(
        config,
        add_config_args,
        "Run module(s) according to configuration specified by a file in `config/`. Note that you can override parts of the configuration on the fly by using the usual CLI flags.",
        "run module(s) in line with a passed configuration file",
    ),
    "prepare": ModuleConfig(
        dataloader.run,
        add_dataloader_args,
        "Run the Data Loader module, to prepare data for use in other modules.",
        "prepare input data",
    ),
    "structure": ModuleConfig(
        structure.run,
        add_structure_args,
        "Run the Structural Discovery module, to learn a structural model for use in training and evaluation.",
        "discover structure",
    ),
    "train": ModuleConfig(
        model.run,
        add_model_args,
        "Run the Architecture module, to train a model.",
        "train a model",
    ),
    "evaluate": ModuleConfig(
        evaluation.run,
        add_evaluation_args,
        "Run the Evaluation module, to evaluate an experiment.",
        "evaluate an experiment",
    ),
    "plot": ModuleConfig(
        plotting.run,
        add_plotting_args,
        "Run the Plotting module, to generate plots for a given model and / or evaluation.",
        "generate plots",
    ),
}



import argparse

from nhssynth.cli.argparse import *
from nhssynth.cli.config import *


def run():

    parser = argparse.ArgumentParser(
        prog="SynthVAE", description="CLI for preparing, training and evaluating a synthetic data generator."
    )
    add_top_level_args(parser)

    # Below we instantiate one subparser for each module
    # + one for running with a config file and one for doing a full run with CLI-specified config
    subparsers = parser.add_subparsers()

    add_module_subparser(subparsers, "pipeline")
    config_parser = add_module_subparser(subparsers, "config")
    add_module_subparser(subparsers, "prepare")
    add_module_subparser(subparsers, "structure")
    add_module_subparser(subparsers, "train")
    add_module_subparser(subparsers, "evaluate")
    add_module_subparser(subparsers, "plot")

    args = parser.parse_args()

    # Use get to return None when no function has been set, i.e. user made no running choice
    executor = vars(args).get("func")

    # If `config` is the specified running choice, we mutate `args` using `parser` and `config_parser` in `read_config`
    # else we execute according to the user's choice
    # else we return `--help` if no choice has been passed, i.e. executor is None
    if executor == config:
        args = read_config(args, parser, config_parser)
    elif executor:
        executor(args)
    else:
        parser.parse_args(["--help"])

    # Whenever either are specified, we want to dump the configuration to allow for this run to be replicated
    if args.save_config or args.save_config_path:
        if not args.save_config_path:
            args.save_config_path = f"config/dump_{args.run_name}.yaml"
        write_config(args)

    print("Complete!")


if __name__ == "__main__":
    run()



from random import gauss
from pandas import Categorical
import torch
import torch.nn as nn

from opacus import PrivacyEngine

# from torch.distributions.bernoulli import Bernoulli
from torch.distributions.normal import Normal

from tqdm import tqdm


class Encoder(nn.Module):
    """Encoder, takes in x
    and outputs mu_z, sigma_z
    (diagonal Gaussian variational posterior assumed)
    """

    def __init__(
        self, input_dim, latent_dim, hidden_dim=32, activation=nn.Tanh, device="gpu",
    ):
        super().__init__()
        if device == "gpu":
            self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
            print(f"Encoder: {device} specified, {self.device} used")
        else:
            self.device = torch.device("cpu")
            print(f"Encoder: {device} specified, {self.device} used")
        output_dim = 2 * latent_dim
        self.latent_dim = latent_dim
        self.net = nn.Sequential(
            nn.Linear(input_dim, hidden_dim),
            activation(),
            nn.Linear(hidden_dim, hidden_dim),
            activation(),
            nn.Linear(hidden_dim, output_dim),
        )

    def forward(self, x):
        outs = self.net(x)
        mu_z = outs[:, : self.latent_dim]
        logsigma_z = outs[:, self.latent_dim :]
        return mu_z, logsigma_z


class Decoder(nn.Module):
    """Decoder, takes in z and outputs reconstruction"""

    def __init__(
        self,
        latent_dim,
        num_continuous,
        num_categories=[0],
        hidden_dim=32,
        activation=nn.Tanh,
        device="gpu",
    ):
        super().__init__()

        output_dim = num_continuous + sum(num_categories)
        self.num_continuous = num_continuous
        self.num_categories = num_categories

        if device == "gpu":
            self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
            print(f"Decoder: {device} specified, {self.device} used")
        else:
            self.device = torch.device("cpu")
            print(f"Decoder: {device} specified, {self.device} used")

        self.net = nn.Sequential(
            nn.Linear(latent_dim, hidden_dim),
            activation(),
            nn.Linear(hidden_dim, hidden_dim),
            activation(),
            nn.Linear(hidden_dim, output_dim),
        )

    def forward(self, z):
        return self.net(z)


class Noiser(nn.Module):
    def __init__(self, num_continuous):
        super().__init__()
        self.output_logsigma_fn = nn.Linear(num_continuous, num_continuous, bias=True)
        torch.nn.init.zeros_(self.output_logsigma_fn.weight)
        torch.nn.init.zeros_(self.output_logsigma_fn.bias)
        self.output_logsigma_fn.weight.requires_grad = False

    def forward(self, X):
        return self.output_logsigma_fn(X)


class VAE(nn.Module):
    """Combines encoder and decoder into full VAE model"""

    def __init__(self, encoder, decoder, lr=1e-3):
        super().__init__()
        self.encoder = encoder.to(encoder.device)
        self.decoder = decoder.to(decoder.device)
        self.device = encoder.device
        self.num_categories = self.decoder.num_categories
        self.num_continuous = self.decoder.num_continuous
        self.noiser = Noiser(self.num_continuous).to(decoder.device)
        self.optimizer = torch.optim.Adam(self.parameters(), lr=lr)
        self.lr = lr

    def reconstruct(self, X):
        mu_z, logsigma_z = self.encoder(X)

        x_recon = self.decoder(mu_z)
        return x_recon

    def generate(self, N):
        z_samples = torch.randn_like(
            torch.ones((N, self.encoder.latent_dim)), device=self.device
        )
        x_gen = self.decoder(z_samples)
        x_gen_ = torch.ones_like(x_gen, device=self.device)
        i = 0

        for v in range(len(self.num_categories)):
            x_gen_[
                :, i : (i + self.num_categories[v])
            ] = torch.distributions.one_hot_categorical.OneHotCategorical(
                logits=x_gen[:, i : (i + self.num_categories[v])]
            ).sample()
            i = i + self.num_categories[v]

        x_gen_[:, -self.num_continuous :] = x_gen[
            :, -self.num_continuous :
        ] + torch.exp(self.noiser(x_gen[:, -self.num_continuous :])) * torch.randn_like(
            x_gen[:, -self.num_continuous :]
        )
        return x_gen_

    def loss(self, X):
        mu_z, logsigma_z = self.encoder(X)

        p = Normal(torch.zeros_like(mu_z), torch.ones_like(mu_z))
        q = Normal(mu_z, torch.exp(logsigma_z))

        divergence_loss = torch.sum(torch.distributions.kl_divergence(q, p))

        s = torch.randn_like(mu_z)
        z_samples = mu_z + s * torch.exp(logsigma_z)

        x_recon = self.decoder(z_samples)

        categoric_loglik = 0
        if sum(self.num_categories) != 0:
            i = 0

            for v in range(len(self.num_categories)):

                categoric_loglik += -torch.nn.functional.cross_entropy(
                    x_recon[:, i : (i + self.num_categories[v])],
                    torch.max(X[:, i : (i + self.num_categories[v])], 1)[1],
                ).sum()
                i = i + self.decoder.num_categories[v]

        gauss_loglik = 0
        if self.decoder.num_continuous != 0:
            gauss_loglik = (
                Normal(
                    loc=x_recon[:, -self.num_continuous :],
                    scale=torch.exp(self.noiser(x_recon[:, -self.num_continuous :])),
                )
                .log_prob(X[:, -self.num_continuous :])
                .sum()
            )

        reconstruct_loss = -(categoric_loglik + gauss_loglik)

        elbo = divergence_loss + reconstruct_loss

        return (elbo, reconstruct_loss, divergence_loss, categoric_loglik, gauss_loglik)

    def train(
        self,
        x_dataloader,
        n_epochs,
        logging_freq=1,
        patience=5,
        delta=10,
        filepath=None,
    ):
        # mean_norm = 0
        # counter = 0
        log_elbo = []
        log_reconstruct = []
        log_divergence = []
        log_cat_loss = []
        log_num_loss = []

        # EARLY STOPPING #
        min_elbo = 0.0  # For early stopping workflow
        patience = patience  # How many epochs patience we give for early stopping
        stop_counter = 0  # Counter for stops
        delta = delta  # Difference in elbo value

        for epoch in range(n_epochs):

            train_loss = 0.0
            divergence_epoch_loss = 0.0
            reconstruction_epoch_loss = 0.0
            categorical_epoch_reconstruct = 0.0
            numerical_epoch_reconstruct = 0.0

            for batch_idx, (Y_subset,) in enumerate(tqdm(x_dataloader)):
                self.optimizer.zero_grad()
                (
                    elbo,
                    reconstruct_loss,
                    divergence_loss,
                    categorical_reconstruc,
                    numerical_reconstruct,
                ) = self.loss(Y_subset.to(self.encoder.device))
                elbo.backward()
                self.optimizer.step()

                train_loss += elbo.item()
                divergence_epoch_loss += divergence_loss.item()
                reconstruction_epoch_loss += reconstruct_loss.item()
                categorical_epoch_reconstruct += categorical_reconstruc.item()
                numerical_epoch_reconstruct += numerical_reconstruct.item()

                # counter += 1
                # l2_norm = 0
                # for p in self.parameters():
                #     if p.requires_grad:
                #         p_norm = p.grad.detach().data.norm(2)
                #         l2_norm += p_norm.item() ** 2
                # l2_norm = l2_norm ** 0.5  # / Y_subset.shape[0]
                # mean_norm = (mean_norm * (counter - 1) + l2_norm) / counter

            log_elbo.append(train_loss)
            log_reconstruct.append(reconstruction_epoch_loss)
            log_divergence.append(divergence_epoch_loss)
            log_cat_loss.append(categorical_epoch_reconstruct)
            log_num_loss.append(numerical_epoch_reconstruct)

            if epoch == 0:

                min_elbo = train_loss

            if train_loss < (min_elbo - delta):

                min_elbo = train_loss
                stop_counter = 0  # Set counter to zero
                if filepath != None:
                    self.save(filepath)  # Save best model if we want to

            else:  # elbo has not improved

                stop_counter += 1

            if epoch % logging_freq == 0:
                print(
                    f"\tEpoch: {epoch:2}. Elbo: {train_loss:11.2f}. Reconstruction Loss: {reconstruction_epoch_loss:11.2f}. KL Divergence: {divergence_epoch_loss:11.2f}. Categorical Loss: {categorical_epoch_reconstruct:11.2f}. Numerical Loss: {numerical_epoch_reconstruct:11.2f}"
                )
                # print(f"\tMean norm: {mean_norm}")
            # self.mean_norm = mean_norm

            if stop_counter == patience:

                n_epochs = epoch + 1

                break

        return (
            n_epochs,
            log_elbo,
            log_reconstruct,
            log_divergence,
            log_cat_loss,
            log_num_loss,
        )

    def diff_priv_train(
        self,
        x_dataloader,
        n_epochs,
        C=1e16,
        noise_scale=None,
        target_eps=1,
        target_delta=1e-5,
        logging_freq=1,
        sample_rate=0.1,
        patience=5,
        delta=10,
        filepath=None,
    ):
        if noise_scale is not None:
            self.privacy_engine = PrivacyEngine(
                self,
                sample_rate=sample_rate,
                alphas=[1 + x / 10.0 for x in range(1, 100)] + list(range(12, 64)),
                noise_multiplier=noise_scale,
                max_grad_norm=C,
            )
        else:
            self.privacy_engine = PrivacyEngine(
                self,
                sample_rate=sample_rate,
                alphas=[1 + x / 10.0 for x in range(1, 100)] + list(range(12, 64)),
                target_epsilon=target_eps,
                target_delta=target_delta,
                epochs=n_epochs,
                max_grad_norm=C,
            )
        self.privacy_engine.attach(self.optimizer)

        log_elbo = []
        log_reconstruct = []
        log_divergence = []
        log_cat_loss = []
        log_num_loss = []

        # EARLY STOPPING #
        min_elbo = 0.0  # For early stopping workflow
        patience = patience  # How many epochs patience we give for early stopping
        stop_counter = 0  # Counter for stops
        delta = delta  # Difference in elbo value

        for epoch in range(n_epochs):
            train_loss = 0.0
            divergence_epoch_loss = 0.0
            reconstruction_epoch_loss = 0.0
            categorical_epoch_reconstruct = 0.0
            numerical_epoch_reconstruct = 0.0
            # print(self.get_privacy_spent(target_delta))

            for batch_idx, (Y_subset,) in enumerate(tqdm(x_dataloader)):

                self.optimizer.zero_grad()
                (
                    elbo,
                    reconstruct_loss,
                    divergence_loss,
                    categorical_reconstruct,
                    numerical_reconstruct,
                ) = self.loss(Y_subset.to(self.encoder.device))
                elbo.backward()
                self.optimizer.step()

                train_loss += elbo.item()
                divergence_epoch_loss += divergence_loss.item()
                reconstruction_epoch_loss += reconstruct_loss.item()
                categorical_epoch_reconstruct += categorical_reconstruct.item()
                numerical_epoch_reconstruct += numerical_reconstruct.item()

                # print(self.get_privacy_spent(target_delta))
                # print(loss.item())

            log_elbo.append(train_loss)
            log_reconstruct.append(reconstruction_epoch_loss)
            log_divergence.append(divergence_epoch_loss)
            log_cat_loss.append(categorical_epoch_reconstruct)
            log_num_loss.append(numerical_epoch_reconstruct)

            if epoch == 0:

                min_elbo = train_loss

            if train_loss < (min_elbo - delta):

                min_elbo = train_loss
                stop_counter = 0  # Set counter to zero
                if filepath != None:
                    self.save(filepath)  # Save best model if we want to

            else:  # elbo has not improved

                stop_counter += 1

            if epoch % logging_freq == 0:
                print(
                    f"\tEpoch: {epoch:2}. Elbo: {train_loss:11.2f}. Reconstruction Loss: {reconstruction_epoch_loss:11.2f}. KL Divergence: {divergence_epoch_loss:11.2f}. Categorical Loss: {categorical_epoch_reconstruct:11.2f}. Numerical Loss: {numerical_epoch_reconstruct:11.2f}"
                )
                # print(f"\tMean norm: {mean_norm}")

            if stop_counter == patience:

                n_epochs = epoch + 1
                break

        return (
            n_epochs,
            log_elbo,
            log_reconstruct,
            log_divergence,
            log_cat_loss,
            log_num_loss,
        )

    def get_privacy_spent(self, delta):
        if hasattr(self, "privacy_engine"):
            return self.privacy_engine.get_privacy_spent(delta)
        else:
            print(
                """This VAE object does not a privacy_engine attribute.
                Run diff_priv_train to create one."""
            )

    def save(self, filename):
        torch.save(self.state_dict(), filename)

    def load(self, filename):
        self.load_state_dict(torch.load(filename))



import argparse
import numpy as np
import pandas as pd
import torch

# For Gower distance
import gower

# For the SUPPORT dataset
from pycox.datasets import support

# VAE functions
from VAE import Decoder, Encoder, VAE

from utils import support_pre_proc, reverse_transformers

# Plotting
import matplotlib

font = {"size": 14}
matplotlib.rc("font", **font)
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable

# For the SUPPORT dataset
from pycox.datasets import support

# VAE functions
from VAE import Decoder, Encoder, VAE

parser = argparse.ArgumentParser()

parser.add_argument(
    "--save_file",
    required=True,
    type=str,
    help="load trained model's state_dict from file",
)

parser.add_argument(
    "--pre_proc_method",
    default="GMM",
    type=str,
    help="Choose the pre-processing method that you will apply to the dataset, either GMM or standard",
)

args = parser.parse_args()

# Import and preprocess the SUPPORT data for ground truth correlations
data_supp = support.read_df()

# Save the original columns

original_continuous_columns = ["duration"] + [f"x{i}" for i in range(7, 15)]
original_categorical_columns = ["event"] + [f"x{i}" for i in range(1, 7)]

original_columns = original_categorical_columns + original_continuous_columns
#%% -------- Data Pre-Processing -------- #
pre_proc_method = args.pre_proc_method

(
    x_train,
    data_supp,
    reordered_dataframe_columns,
    continuous_transformers,
    categorical_transformers,
    num_categories,
    num_continuous,
) = support_pre_proc(data_supp=data_supp, pre_proc_method=pre_proc_method)


###############################################################################

# Load saved model - ensure parameters are equivalent to the saved model
latent_dim = 256
hidden_dim = 256
encoder = Encoder(x_train.shape[1], latent_dim, hidden_dim=hidden_dim)
decoder = Decoder(latent_dim, num_continuous, num_categories=num_categories)
vae = VAE(encoder, decoder)
vae.load(args.save_file)

#%% -------- Generate Synthetic Data -------- #

# Generate a synthetic set using trained vae

synthetic_trial = vae.generate(data_supp.shape[0])  # 8873 is size of support
#%% -------- Inverse Transformation On Synthetic Trial -------- #

synthetic_sample = vae.generate(data_supp.shape[0])

if torch.cuda.is_available():
    synthetic_sample = pd.DataFrame(
        synthetic_sample.cpu().detach(), columns=reordered_dataframe_columns
    )
else:
    synthetic_sample = pd.DataFrame(
        synthetic_sample.detach(), columns=reordered_dataframe_columns
    )

# Reverse the transformations

synthetic_supp = reverse_transformers(
    synthetic_set=synthetic_sample,
    data_supp_columns=data_supp.columns,
    cont_transformers=continuous_transformers,
    cat_transformers=categorical_transformers,
    pre_proc_method=pre_proc_method,
)


### Create plots
# Plot 1: Correlation matrix of original data
plt.figure()
ax = plt.gca()
ax.axes.get_xaxis().set_visible(False)
ax.axes.get_yaxis().set_visible(False)
im = ax.matshow(data_supp.corr())
#####
# Credit: https://stackoverflow.com/questions/18195758/set-matplotlib-colorbar-size-to-match-graph
divider = make_axes_locatable(ax)
cax = divider.append_axes("right", size="5%", pad=0.05)
#####
plt.colorbar(im, cax=cax)
plt.savefig("actual_corr_{}.png".format(pre_proc_method), bbox_inches="tight")
# Plot 2: Correlation matrix of synthetic data
plt.figure()
ax = plt.gca()
ax.axes.get_xaxis().set_visible(False)
ax.axes.get_yaxis().set_visible(False)
im = ax.matshow(synthetic_supp.corr())
#####
# Credit: https://stackoverflow.com/questions/18195758/set-matplotlib-colorbar-size-to-match-graph
divider = make_axes_locatable(ax)
cax = divider.append_axes("right", size="5%", pad=0.05)
#####
plt.colorbar(im, cax=cax)
plt.savefig("sample_corr_{}.png".format(pre_proc_method), bbox_inches="tight")
# Plot 3: Difference between real and synth correlation matrices + Gower and RMSE values
plt.figure()
g = np.mean(gower.gower_matrix(data_supp, synthetic_supp))
p = np.sqrt(
    np.mean((data_supp.corr().to_numpy() - synthetic_supp.corr().to_numpy()) ** 2)
)
plt.title(f"Gower Distance = {g:.4f}\n Correlation RMSE = {p:.4f}")
ax = plt.gca()
ax.axes.get_xaxis().set_visible(False)
ax.axes.get_yaxis().set_visible(False)
im = ax.matshow(synthetic_supp.corr() - data_supp.corr())
#####
# Credit: https://stackoverflow.com/questions/18195758/set-matplotlib-colorbar-size-to-match-graph
divider = make_axes_locatable(ax)
cax = divider.append_axes("right", size="5%", pad=0.05)
#####
plt.colorbar(im, cax=cax)
plt.savefig("diff_corr_{}.png".format(pre_proc_method), bbox_inches="tight")



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import torch
from rdt.transformers import categorical, datetime, numerical
from sklearn.preprocessing import StandardScaler


def set_seed(seed):
    np.random.seed(seed)
    torch.manual_seed(seed)


# -------- Pre-Processing for SUPPORT -------- #

gmm_seed = 0


def support_pre_proc(data_supp, pre_proc_method="GMM"):

    #%% -------- Data Pre-Processing -------- #

    # We one-hot the categorical cols and standardise the continuous cols
    data_supp["x14"] = data_supp["x0"]
    # data_supp = data_supp.astype('float32')
    data_supp = data_supp[["duration"] + [f"x{i}" for i in range(1, 15)] + ["event"]]
    data_supp[["x1", "x2", "x3", "x4", "x5", "x6", "event"]] = data_supp[
        ["x1", "x2", "x3", "x4", "x5", "x6", "event"]
    ].astype(int)

    # As of coding this, new version of RDT adds in GMM transformer which is what we require, however hyper transformers do not work as individual
    # transformers take a 'columns' argument that can only allow for fitting of one column - so you need to loop over and create one for each column
    # in order to fit the dataset - https://github.com/sdv-dev/RDT/issues/376

    continuous_transformers = {}
    categorical_transformers = {}

    continuous_columns = ["duration"] + [f"x{i}" for i in range(7, 15)]
    categorical_columns = ["event"] + [f"x{i}" for i in range(1, 7)]
    num_categories = (np.array([np.amax(data_supp[col]) for col in categorical_columns]) + 1).astype(int)
    num_continuous = len(continuous_columns)

    transformed_dataset = data_supp.copy(deep=True)

    # Define columns based on datatype and then loop over creating and fitting
    # transformers

    if pre_proc_method == "GMM":
        for index, column in enumerate(continuous_columns):
            # Fit GMM
            temp_continuous = numerical.ClusterBasedNormalizer()
            temp_continuous.fit(transformed_dataset, column=column)
            continuous_transformers["continuous_{}".format(column)] = temp_continuous

            transformed_dataset = temp_continuous.transform(transformed_dataset)

            # Each numerical one gets a .normalized column + a .component column giving the mixture info
            # This too needs to be one hot encoded

            categorical_columns += [str(column) + ".component"]

            # Let's retrieve the new categorical and continuous column names

            continuous_columns = ["duration.normalized"] + [f"x{i}.normalized" for i in range(7, 15)]

            # For each categorical column we want to know the number of categories

            num_categories = (np.array([np.amax(transformed_dataset[col]) for col in categorical_columns]) + 1).astype(
                int
            )

            num_continuous = len(continuous_columns)

    elif pre_proc_method == "standard":
        for index, column in enumerate(continuous_columns):
            # Fit sklearn standard scaler to each column
            temp_continuous = StandardScaler()
            temp_column = transformed_dataset[column].values.reshape(-1, 1)
            temp_continuous.fit(temp_column)
            continuous_transformers["continuous_{}".format(column)] = temp_continuous

            transformed_dataset[column] = (temp_continuous.transform(temp_column)).flatten()

    for index, column in enumerate(categorical_columns):

        temp_categorical = categorical.OneHotEncoder()
        temp_categorical.fit(transformed_dataset, column=column)
        categorical_transformers["categorical_{}".format(index)] = temp_categorical

        transformed_dataset = temp_categorical.transform(transformed_dataset)

    # We need the dataframe in the correct format i.e. categorical variables first and in the order of
    # num_categories with continuous variables placed after

    reordered_dataframe = pd.DataFrame()

    reordered_dataframe = transformed_dataset.iloc[:, num_continuous:]

    reordered_dataframe = pd.concat(
        [reordered_dataframe, transformed_dataset.iloc[:, :num_continuous]],
        axis=1,
    )

    x_train_df = reordered_dataframe.to_numpy()
    x_train = x_train_df.astype("float32")

    return (
        x_train,
        data_supp,
        reordered_dataframe.columns,
        continuous_transformers,
        categorical_transformers,
        num_categories,
        num_continuous,
    )


# -------- Pre-Processing for MIMIC sets -------- #
# Internal sets provided by NHSX - outside users will have to stick with SUPPORT set


def mimic_pre_proc(data_supp, pre_proc_method="GMM"):

    # Specify column configurations

    original_categorical_columns = [
        "ETHNICITY",
        "DISCHARGE_LOCATION",
        "GENDER",
        "FIRST_CAREUNIT",
        "VALUEUOM",
        "LABEL",
    ]
    original_continuous_columns = ["SUBJECT_ID", "VALUE", "age"]
    original_datetime_columns = ["ADMITTIME", "DISCHTIME", "DOB", "CHARTTIME"]

    categorical_columns = original_categorical_columns.copy()
    continuous_columns = original_continuous_columns.copy()
    datetime_columns = original_datetime_columns.copy()

    # As of coding this, new version of RDT adds in GMM transformer which is what we require, however hyper transformers do not work as individual
    # transformers take a 'columns' argument that can only allow for fitting of one column - so you need to loop over and create one for each column
    # in order to fit the dataset - https://github.com/sdv-dev/RDT/issues/376

    continuous_transformers = {}
    categorical_transformers = {}
    datetime_transformers = {}

    # num_categories is either the maximum number within the categorical
    # column, or the number of unique string terms

    num_categories = []

    for col in original_categorical_columns:

        if data_supp[col].dtype == float:

            # Convert to int
            data_supp[col] = data_supp[col].astype(int)

        if data_supp[col].dtype == int:

            num_categories.append(np.amax(data_supp[col]) + 1)

        # Categories are strings
        else:

            # Convert column into one type
            values = np.unique(data_supp[col].astype(str), return_counts=False)
            num_categories.append(values.shape[0])

    num_continuous = len(original_continuous_columns)

    transformed_dataset = data_supp.copy(deep=True)

    # Define columns based on datatype and then loop over creating and fitting
    # transformers

    # Do datetime columns first to convert to seconds

    for index, column in enumerate(original_datetime_columns):

        # Fit datetime transformer - converts to seconds
        temp_datetime = datetime.OptimizedTimestampEncoder()
        temp_datetime.fit(transformed_dataset, column=column)
        datetime_transformers["datetime_{}".format(column)] = temp_datetime

        transformed_dataset = temp_datetime.transform(transformed_dataset)
        # These newly fitted datetime columns now need to be scaled
        # And treated as a continuous variable
        continuous_columns += [str(column) + ".value"]

    # WE NEED TO RETAIN THIS SET AS METRICS DO NOT EVALUATE WITH DATETIMES BUT THEY WILL EVALUATE
    # IF DATETIMES ARE IN A SECONDS FORMAT

    original_metric_set = transformed_dataset.copy(deep=True)

    if pre_proc_method == "GMM":

        for index, column in enumerate(continuous_columns):

            # Fit GMM
            temp_continuous = numerical.BayesGMMTransformer()
            temp_continuous.fit(transformed_dataset, column=column)
            continuous_transformers["continuous_{}".format(column)] = temp_continuous

            categorical_columns += [str(column) + ".component"]

            transformed_dataset = temp_continuous.transform(transformed_dataset)

        # Each numerical one gets a .normalized column + a .component column giving the mixture info
        # This too needs to be one hot encoded

        continuous_columns = [str(col) + ".normalized" for col in continuous_columns]

    elif pre_proc_method == "standard":

        for index, column in enumerate(continuous_columns):

            # Fit sklearn standard scaler to each column
            temp_continuous = StandardScaler()
            temp_column = transformed_dataset[column].values.reshape(-1, 1)
            temp_continuous.fit(temp_column)
            continuous_transformers["continuous_{}".format(column)] = temp_continuous

            transformed_dataset[column] = (temp_continuous.transform(temp_column)).flatten()

    num_categories = []

    for col in categorical_columns:

        if transformed_dataset[col].dtype == float:

            # Convert to int
            transformed_dataset[col] = transformed_dataset[col].astype(int)

        if transformed_dataset[col].dtype == int:

            num_categories.append(np.amax(transformed_dataset[col]) + 1)

        # Categories are strings/objects
        else:

            # Convert column into one type
            values = np.unique(transformed_dataset[col].astype(str), return_counts=False)

            num_categories.append(values.shape[0])

    num_continuous = len(continuous_columns)

    for index, column in enumerate(categorical_columns):

        temp_categorical = categorical.OneHotEncoder()
        temp_categorical.fit(transformed_dataset, column=column)
        categorical_transformers["categorical_{}".format(index)] = temp_categorical

        transformed_dataset = temp_categorical.transform(transformed_dataset)

    # We need the dataframe in the correct format i.e. categorical variables first and in the order of
    # num_categories with continuous variables placed after

    reordered_dataframe = pd.DataFrame()

    reordered_dataframe = transformed_dataset.iloc[:, num_continuous:]

    reordered_dataframe = pd.concat(
        [reordered_dataframe, transformed_dataset.iloc[:, :num_continuous]],
        axis=1,
    )

    x_train_df = reordered_dataframe.to_numpy()
    x_train = x_train_df.astype("float32")

    return (
        x_train,
        original_metric_set,
        reordered_dataframe.columns,
        continuous_transformers,
        categorical_transformers,
        datetime_transformers,
        num_categories,
        num_continuous,
    )


# -------- Reverse Transformations -------- #


def reverse_transformers(
    synthetic_set,
    data_supp_columns,
    cont_transformers=None,
    cat_transformers=None,
    date_transformers=None,
    pre_proc_method="GMM",
):

    # Now all of the transformations from the dictionary - first loop over the categorical columns

    synthetic_transformed_set = synthetic_set.copy(deep=True)

    if cat_transformers != None:
        for transformer_name in cat_transformers:

            transformer = cat_transformers[transformer_name]
            column_name = transformer_name[12:]

            synthetic_transformed_set = transformer.reverse_transform(synthetic_transformed_set)

    if cont_transformers != None:

        if pre_proc_method == "GMM":

            for transformer_name in cont_transformers:

                transformer = cont_transformers[transformer_name]
                column_name = transformer_name[11:]

                synthetic_transformed_set = transformer.reverse_transform(synthetic_transformed_set)

        elif pre_proc_method == "standard":

            for transformer_name in cont_transformers:

                transformer = cont_transformers[transformer_name]
                column_name = transformer_name[11:]

                # Reverse the standard scaling
                synthetic_transformed_set[column_name] = transformer.inverse_transform(
                    synthetic_transformed_set[column_name].values.reshape(-1, 1)
                ).flatten()

    if date_transformers != None:
        for transformer_name in date_transformers:

            transformer = date_transformers[transformer_name]
            column_name = transformer_name[9:]

            synthetic_transformed_set = transformer.reverse_transform(synthetic_transformed_set)

    synthetic_transformed_set = pd.DataFrame(synthetic_transformed_set, columns=data_supp_columns)

    return synthetic_transformed_set


# -------- Constraint based sampling for MIMIC work -------- #


def constraint_filtering(
    n_rows,
    vae,
    reordered_cols,
    data_supp_columns,
    cont_transformers,
    cat_transformers,
    date_transformers,
    reverse_transformers=reverse_transformers,
    pre_proc_method="GMM",
):

    # Generate samples
    synthetic_trial = vae.generate(n_rows)

    if torch.cuda.is_available():
        # Create pandas dataframe in column order
        synthetic_dataframe = pd.DataFrame(synthetic_trial.cpu().detach().numpy(), columns=reordered_cols)
    else:
        # Create pandas dataframe in column order
        synthetic_dataframe = pd.DataFrame(synthetic_trial.detach().numpy(), columns=reordered_cols)

    # Reverse all the transformations ready for filtering
    synthetic_dataframe = reverse_transformers(
        synthetic_dataframe,
        data_supp_columns,
        cont_transformers,
        cat_transformers,
        date_transformers,
        pre_proc_method=pre_proc_method,
    )

    # Function to filter out the constraints from the set - returns valid dataframe
    def constraint_check(synthetic_df):
        # age greater than 0                   patient was discharged after being admitted           patient admitted after their date of birth         patient first chart after admit time
        valid_df = synthetic_df[
            (synthetic_df["age"] > 0)
            | (synthetic_df["DISCHTIME"] >= synthetic_df["ADMITTIME"])
            | (synthetic_df["ADMITTIME"] > synthetic_df["DOB"])
            | (synthetic_df["CHARTTIME"] >= synthetic_df["ADMITTIME"])
        ]

        return valid_df

    # Do first check
    synthetic_dataframe = constraint_check(synthetic_dataframe)

    # Loop over returning a valid dataframe each time until we get a set that is big enough
    while synthetic_dataframe.shape[0] != n_rows:

        rows_needed = n_rows - synthetic_dataframe.shape[0]

        # If we have too many, remove the required amount
        if rows_needed < 0:

            rows_needed = np.arange(abs(rows_needed))

            # Drop the bottom rows_needed amount

            synthetic_dataframe.drop(rows_needed, axis=0, inplace=True)

        # Need to generate enough to fill the dataframe
        else:

            new_set = vae.generate(rows_needed)

            new_set = pd.DataFrame(new_set.cpu().detach().numpy(), columns=reordered_cols)

            new_set = reverse_transformers(
                new_set,
                data_supp_columns,
                cont_transformers,
                cat_transformers,
                date_transformers,
            )

            new_filtered_set = constraint_check(new_set)

            # Add this onto the original and re-run
            synthetic_dataframe = pd.concat([synthetic_dataframe, new_filtered_set])

    return synthetic_dataframe


def plot_elbo(
    n_epochs,
    log_elbo,
    log_reconstruction,
    log_divergence,
    saving_filepath=None,
    pre_proc_method="GMM",
):

    x = np.arange(n_epochs)

    y1 = log_elbo
    y2 = log_reconstruction
    y3 = log_divergence

    plt.plot(x, y1, label="ELBO")
    plt.plot(x, y2, label="RECONSTRUCTION")
    plt.plot(x, y3, label="DIVERGENCE")
    plt.xlabel("Number of Epochs")
    # Set the y axis label of the current axis.
    plt.ylabel("Loss Value")
    # Set a title of the current axes.
    plt.title("ELBO Breakdown")
    # show a legend on the plot
    plt.legend()

    if saving_filepath != None:
        # Save static image
        plt.savefig("{}ELBO_Breakdown_SynthVAE_{}.png".format(saving_filepath, pre_proc_method))

    plt.show()

    return None


def plot_likelihood_breakdown(
    n_epochs,
    log_categorical,
    log_numerical,
    saving_filepath=None,
    pre_proc_method="GMM",
):

    x = np.arange(n_epochs)

    y1 = log_categorical
    y2 = log_numerical

    plt.subplot(1, 2, 1)
    plt.plot(x, y1, label="CATEGORICAL")
    plt.xlabel("Number of Epochs")
    # Set the y axis label of the current axis.
    plt.ylabel("Loss Value")
    # Set a title of the current axes.
    plt.title("Categorical Breakdown")
    # show a legend on the plot
    plt.subplot(1, 2, 2)
    plt.plot(x, y2, label="NUMERICAL")
    plt.xlabel("Number of Epochs")
    # Set the y axis label of the current axis.
    plt.ylabel("Loss Value")
    # Set a title of the current axes.
    plt.title("Numerical Breakdown")
    # show a legend on the plot
    plt.tight_layout()

    if saving_filepath != None:
        # Save static image
        plt.savefig("{}Reconstruction_Breakdown_SynthVAE_{}.png".format(saving_filepath, pre_proc_method))

    return None


def plot_variable_distributions(
    categorical_columns,
    continuous_columns,
    data_supp,
    synthetic_supp,
    saving_filepath=None,
    pre_proc_method="GMM",
):

    # Plot some examples using plotly

    for column in categorical_columns:

        plt.subplot(1, 2, 1)
        plt.hist(x=synthetic_supp[column])
        plt.title("Synthetic")
        # Set the x axis label of the current axis
        plt.xlabel("Data Value")
        # Set the y axis label of the current axis.
        plt.ylabel("Distribution")
        # Set a title of the current axes.
        plt.title("Synthetic".format(column))
        # show a legend on the plot
        plt.subplot(1, 2, 2)
        plt.hist(x=data_supp[column])
        plt.title("Original")
        # Set the x axis label of the current axis
        plt.xlabel("Data Value")
        # Set the y axis label of the current axis.
        plt.ylabel("Distribution")
        # Set a title of the current axes.
        plt.title("Original".format(column))
        # show a legend on the plot
        plt.suptitle("Variable {}".format(column))

        plt.tight_layout()

        if saving_filepath != None:
            # Save static image
            plt.savefig("{}Variable_{}_SynthVAE_{}.png".format(saving_filepath, column, pre_proc_method))

        plt.show()

    for column in continuous_columns:

        plt.subplot(1, 2, 1)
        plt.hist(x=synthetic_supp[column])
        plt.title("Synthetic")
        # Set the x axis label of the current axis
        plt.xlabel("Data Value")
        # Set the y axis label of the current axis.
        plt.ylabel("Distribution")
        # Set a title of the current axes.
        plt.title("Synthetic".format(column))
        # show a legend on the plot
        plt.subplot(1, 2, 2)
        plt.hist(x=data_supp[column])
        plt.title("Original")
        # Set the x axis label of the current axis
        plt.xlabel("Data Value")
        # Set the y axis label of the current axis.
        plt.ylabel("Distribution")
        # Set a title of the current axes.
        plt.title("Original".format(column))
        # show a legend on the plot
        plt.suptitle("Variable {}".format(column))

        plt.tight_layout()

        if saving_filepath != None:
            # Save static image
            plt.savefig("{}Variable_{}_SynthVAE_{}.png".format(saving_filepath, column, pre_proc_method))

        plt.show()

        return None



import numpy as np
import pandas as pd
from nhssynth.modules.dataloader.io import *
from nhssynth.modules.dataloader.metadata import *
from nhssynth.modules.dataloader.transformers import *


def run(args) -> None:

    np.random.seed(args.seed)

    input_path, output_path, metadata_input_path, metadata_output_path, experiment_path = format_io(
        args.input_file, args.output_file, args.metadata_file, args.dir, run_name=args.run_name
    )
    experiment_path.mkdir(parents=True, exist_ok=True)

    input = pd.read_csv(input_path, index_col=args.index_col)
    metadata = load_metadata(metadata_input_path, input)

    dtypes = instantiate_dtypes(metadata, input)
    # TODO point out when this fails that it must be due to an invalid / unsupported dtype in the metadata
    typed_input = input.astype(dtypes)

    metatransformer = instantiate_metatransformer(
        metadata, typed_input, args.sdv_workflow, args.allow_null_transformers
    )

    output_metadata(metadata_output_path, dtypes, metatransformer, args.sdv_workflow, args.collapse_yaml)

    transformed_input = apply_transformer(metatransformer, typed_input)



import argparse

import nhssynth.cli.module_setup as ms
import yaml


def get_default_args(parser: argparse.ArgumentParser):
    defaults = {}
    for action in parser._actions:
        if action.dest not in ["help", "==SUPPRESS=="]:
            defaults[action.dest] = action.default
    return defaults


def read_config(
    args: argparse.Namespace,
    parser: argparse.ArgumentParser,
    config_parser: argparse.ArgumentParser,
) -> argparse.Namespace:

    # Get all possible default arguments by scraping the top level `parser` and the `config_parser`
    args_dict = get_default_args(parser)
    args_dict.update(get_default_args(config_parser))

    # Find the non-default arguments amongst passed `args` by seeing which of them are different to the entries of `args_dict`
    non_default_passed_args_dict = {k: v for k, v in vars(args).items() if k != "func" and v != args_dict[k]}

    # Open the passed yaml file and load into a dictionary
    with open(f"config/{args.input_config}.yaml") as stream:
        config_args_dict = yaml.safe_load(stream)

    # Overwrite the default arguments with the ones from the yaml file
    args_dict.update(config_args_dict)
    # Overwrite the result of the above with any non-default CLI args
    args_dict.update(non_default_passed_args_dict)

    # Create a new Namespace using the assembled dictionary
    new_args = argparse.Namespace(**args_dict)

    # Run the appropriate execution function
    new_args.func = ms.MODULE_MAP[new_args.run_type].func
    new_args.func(new_args)
    return new_args


def write_config(args: argparse.Namespace):

    # TODO Tidy up the yaml and make it hierarchical rather than one big list
    args_dict = vars(args)
    args_dict["run_type"] = next(
        (run_type for run_type, mc in ms.MODULE_MAP.items() if mc.func == args_dict["func"]), None
    )
    with open(f"{args.save_config_path}", "w") as yaml_file:
        del args_dict["func"], args_dict["run_name"], args_dict["save_config"], args_dict["save_config_path"]
        yaml.dump(args_dict, yaml_file, default_flow_style=False)



import argparse


def add_all_module_args(parser: argparse.ArgumentParser):
    dataloader_group = parser.add_argument_group(title="dataloader")
    add_dataloader_args(dataloader_group)
    structure_group = parser.add_argument_group(title="structure")
    add_structure_args(structure_group)
    model_group = parser.add_argument_group(title="model")
    add_model_args(model_group)
    evaluation_group = parser.add_argument_group(title="evaluation")
    add_evaluation_args(evaluation_group)
    plotting_group = parser.add_argument_group(title="plotting")
    add_plotting_args(plotting_group)


def add_config_args(parser: argparse.ArgumentParser):
    parser.add_argument(
        "--input-config",
        "-c",
        required=True,
        help="Specify the config file.",
    )
    overrides_group = parser.add_argument_group(title="overrides")
    # TODO is there a way to do this using `add_all_module_args`, i.e. can we nest groups? Doesn't seem to work
    add_dataloader_args(overrides_group, override=True)
    add_structure_args(overrides_group, override=True)
    add_model_args(overrides_group, override=True)
    add_evaluation_args(overrides_group, override=True)
    add_plotting_args(overrides_group, override=True)


def add_dataloader_args(parser: argparse.ArgumentParser, override=False):
    parser.add_argument(
        "-i",
        "--input-file",
        required=(not override),
        help="the name of the `.csv` file to prepare",
    )
    parser.add_argument(
        "-o",
        "--output-file",
        default="_prepared",
        help="where to write the prepared data, defaults to `experiments/<args.run_name>/<args.input_file>_prepared.csv`",
    )
    parser.add_argument(
        "-d",
        "--dir",
        default="./data",
        help="the directory to read and write data from and to",
    )
    parser.add_argument(
        "-m",
        "--metadata-file",
        default="_metadata",
        help="metadata for the input data, defaults to `<args.dir>/<args.input_file>_metadata.yaml`",
    )
    parser.add_argument(
        "-ic",
        "--index-col",
        default=None,
        choices=[None, 0],
        help="indicate whether the csv file's 0th column is an index column, such that pandas can ignore it",
    )
    parser.add_argument(
        "-sdv",
        "--sdv-workflow",
        action="store_true",
        help="utilise the SDV synthesizer workflow for transformation and metadata, rather than a `HyperTransformer` from RDT",
    )
    parser.add_argument(
        "-ant",
        "--allow-null-transformers",
        action="store_true",
        help="allow null / None transformers, i.e. leave some columns as they are",
    )
    parser.add_argument(
        "-cy",
        "--collapse-yaml",
        action="store_true",
        help="use aliases and anchors in the output metadata yaml, this will make it much more compact",
    )
    parser.add_argument(
        "-is",
        "--imputation-strategy",
        default="mean",
        choices=["mean", "median", "cull"],
        help="imputation strategy for missing values",
    )


def add_structure_args(parser: argparse.ArgumentParser, override=False):
    pass


def add_model_args(parser: argparse.ArgumentParser, override=False):
    pass


def add_evaluation_args(parser: argparse.ArgumentParser, override=False):
    pass


def add_plotting_args(parser: argparse.ArgumentParser, override=False):
    pass



import pathlib
import warnings

import numpy as np
import pandas as pd
import yaml
from nhssynth.modules.dataloader.utils import *


def create_empty_metadata(data: pd.DataFrame) -> dict[str, dict]:
    return {cn: {} for cn in data.columns}


def check_metadata_columns(metadata: dict, data: pd.DataFrame):
    assert all([k in data.columns for k in metadata.keys()])
    metadata.update({cn: {} for cn in data.columns if cn not in metadata})


def load_metadata(in_path: pathlib.Path, data: pd.DataFrame):
    if in_path.exists():
        with open(in_path) as stream:
            metadata = yaml.safe_load(stream)
        metadata.pop("transformers", None)
        metadata.pop("column_types", None)
    else:
        metadata = create_empty_metadata(data)
    check_metadata_columns(metadata, data)
    return metadata


def instantiate_dtypes(metadata: dict, data: pd.DataFrame) -> dict[str, np.dtype]:
    dtypes = {cn: cd.get("dtype", {}) for cn, cd in metadata.items()}
    if not all(dtypes.values()):
        warnings.warn(
            f"Incomplete metadata, detecting missing `dtype`s for column(s): {[k for k, v in dtypes.items() if not v]} automatically...",
            UserWarning,
        )
        dtypes.update({cn: data[cn].dtype for cn, cv in dtypes.items() if not cv})
    return dtypes


def assemble_metadata(dtypes, metatransformer, sdv_workflow):
    if sdv_workflow:
        sdmetadata, synthesizer = metatransformer
        transformers = synthesizer.get_transformers()
        metadata = {
            cn: {
                **cd,
                **{
                    "transformer": {
                        "name": type(transformers[cn]).__name__,
                        **filter_inner_dict(
                            transformers[cn].__dict__,
                            {"output_properties", "random_states", "transform", "reverse_transform", "_dtype"},
                        ),
                    }
                    if transformers[cn]
                    else None,
                    "dtype": dtypes[cn].name if not isinstance(dtypes[cn], str) else dtypes[cn],
                },
            }
            for cn, cd in sdmetadata.columns.items()
        }
    else:
        config = metatransformer.get_config()
        metadata = {
            cn: {
                "sdtype": cd,
                "transformer": {
                    "name": type(config["transformers"][cn]).__name__,
                    **filter_inner_dict(
                        config["transformers"][cn].__dict__,
                        {"output_properties", "random_states", "transform", "reverse_transform", "_dtype"},
                    ),
                }
                if config["transformers"][cn]
                else None,
                "dtype": dtypes[cn].name if not isinstance(dtypes[cn], str) else dtypes[cn],
            }
            for cn, cd in config["sdtypes"].items()
        }
    return metadata


def get_key_by_value(dict, value):
    for key, val in dict.items():
        if val == value:
            return key
    return None


def output_metadata(out_path: pathlib.Path, dtypes, metatransformer, sdv_workflow=True, collapse_yaml=True):

    metadata = assemble_metadata(dtypes, metatransformer, sdv_workflow)

    if collapse_yaml:
        c_index = 1
        column_types = {}
        t_index = 1
        transformers = {}
        for cn, cd in metadata.items():
            if cd not in column_types.values():
                column_types[c_index] = cd.copy()
                metadata[cn] = column_types[c_index]
                c_index += 1
            else:
                cix = get_key_by_value(column_types, cd)
                metadata[cn] = column_types[cix]
            if cd["transformer"] not in transformers.values() and cd["transformer"]:
                transformers[t_index] = cd["transformer"].copy()
                metadata[cn]["transformer"] = transformers[t_index]
                t_index += 1
            elif cd["transformer"]:
                tix = get_key_by_value(transformers, cd["transformer"])
                metadata[cn]["transformer"] = transformers[tix]
        metadata = {"transformers": transformers, "column_types": column_types, **metadata}

    with open(out_path, "w") as yaml_file:
        yaml.safe_dump(metadata, yaml_file, default_flow_style=False, sort_keys=False)
